# dualPredictor/df_preprocess.py

import pandas as pd
from sklearn.preprocessing import LabelEncoder, StandardScaler
from sklearn.impute import KNNImputer

def check_df_forms(df_train, df_test):
    # Check if the DataFrames have the same columns in the same order
    if list(df_train.columns) == list(df_test.columns):
        print("Both DataFrames have the same columns in the same order.")
        return df_train, df_test

    # Find the overlap of columns between df_train and df_test
    common_cols = df_train.columns.intersection(df_test.columns)

    # Select the overlap of columns for both DataFrames
    df_train_processed = df_train[common_cols]
    df_test_processed = df_test[common_cols]

    # Print the number of columns kept and dropped for each DataFrame
    print(f"df_train: Kept {len(common_cols)} columns, Dropped {len(df_train.columns) - len(common_cols)} columns.")
    print(f"df_test: Kept {len(common_cols)} columns, Dropped {len(df_test.columns) - len(common_cols)} columns.")

    return df_train_processed, df_test_processed

# Example usage:
# df_train, df_test=check_df_forms(df_train, df_test)
# dualPredictor/df_preprocess.py

import pandas as pd
from sklearn.preprocessing import LabelEncoder, StandardScaler
from sklearn.impute import KNNImputer

def check_df_forms(df_train, df_test):
    # Check if the DataFrames have the same columns in the same order
    if list(df_train.columns) == list(df_test.columns):
        print("Both DataFrames have the same columns in the same order.")
        return df_train, df_test

    # Find the overlap of columns between df_train and df_test
    common_cols = df_train.columns.intersection(df_test.columns)

    # Select the overlap of columns for both DataFrames
    df_train_processed = df_train[common_cols]
    df_test_processed = df_test[common_cols]

    # Print the number of columns kept and dropped for each DataFrame
    print(f"df_train: Kept {len(common_cols)} columns, Dropped {len(df_train.columns) - len(common_cols)} columns.")
    print(f"df_test: Kept {len(common_cols)} columns, Dropped {len(df_test.columns) - len(common_cols)} columns.")

    return df_train_processed, df_test_processed

# Example usage:
# df_train, df_test=check_df_forms(df_train, df_test)

def data_preprocessing(df, target_col, id_col=None, drop_cols=None, scaler=None, imputer=None):
    # Drop specified columns
    if drop_cols is not None:
        df = df.drop(columns=drop_cols)

    # Drop rows with missing values in the target column if this is a df_train (scaler = None)
    if scaler is None:
      df = df.dropna(subset=[target_col])

    # Set id_col as the index column
    if id_col is not None:
        df = df.set_index(id_col)

    # Split the data into features and target
    X = df.drop(target_col, axis=1)
    y = df[target_col]

    # display the stats
    # print('Before the pre-processing')
    # display(X.describe())
    # display(y.describe())

    # Detect numerical and categorical columns
    num_cols = X.select_dtypes(include=['int64', 'float64']).columns.tolist()
    cat_cols = X.select_dtypes(include=['object']).columns.tolist()

    # Label encode categorical features
    for col in cat_cols:
        le = LabelEncoder()
        X[col] = le.fit_transform(X[col])
    
    # Scale numerical features
    if scaler is None:
        scaler = StandardScaler()
        X[num_cols] = scaler.fit_transform(X[num_cols])
    else:
        X[num_cols] = scaler.transform(X[num_cols])
    

    # Impute missing values
    if imputer is None:
        imputer = KNNImputer(n_neighbors=2,keep_empty_features=True)
        X = pd.DataFrame(imputer.fit_transform(X), columns=X.columns)
    
        # display(X.shape)
    else:
        X = pd.DataFrame(imputer.fit_transform(X), columns=X.columns)

    X = X.set_index(df.index)

    return X, y,scaler,imputer



# dualPredictor/dual_model.py

from sklearn.base import BaseEstimator, RegressorMixin, clone
from sklearn.linear_model import LassoCV, RidgeCV, LinearRegression
from sklearn.metrics import confusion_matrix, balanced_accuracy_score,fbeta_score
import numpy as np

class DualModel(BaseEstimator, RegressorMixin):
    def __init__(self, model_type='lasso', metric='youden_index',default_cut_off=0.5):
        self.model_type = model_type
        self.metric=metric
        self.default_cut_off = default_cut_off
        self.model = None
        self.optimal_cut_off = None
        self.y_label_true_=None
        self.metrics=None
        self.cutoffs=None

        if model_type == 'lasso':
            self.model = LassoCV(cv=5)
        elif model_type == 'ridge':
            self.model = RidgeCV(cv=5)
        elif model_type == 'ols':
            self.model = LinearRegression()
            self.model.alpha_ = 0  # Set alpha to 0 for OLS
        else:
            raise ValueError("Unsupported model type. Choose 'lasso', 'ridge', or 'ols'.")

    def fit(self, X, y):
        self.model.fit(X, y)

        # Create binary labels based on the default cut-off
        y_label_true = (y < self.default_cut_off).astype(int)
        self.y_label_true_=y_label_true
        # error hadnling for the user input default cut_off
        if self.default_cut_off >= np.max(y):
                    raise ValueError("The default cut-off must be smaller than the maximum value of y.")

        # Tune the optimal cut-off
        cut_offs = np.linspace(self.default_cut_off, max(y), 55)
        metrics = []

        for cut_off in cut_offs:
            y_pred = (self.model.predict(X) < cut_off).astype(int)
            if self.metric == 'f1_score':
                metric_value = fbeta_score(y_label_true, y_pred,beta=1)
            elif self.metric == 'f2_score':
                metric_value = fbeta_score(y_label_true, y_pred,beta=2)
            elif self.metric == 'youden_index':
                metric_value = balanced_accuracy_score(y_label_true, y_pred, adjusted=True)
            else:
                raise ValueError("Unsupported metric. Choose 'f1_score', 'f2_score', or 'youden_index'.")

            metrics.append(metric_value)

        max_metric = max(metrics)
        max_indices = [i for i, x in enumerate(metrics) if x == max_metric]
        middle_index = max_indices[len(max_indices) // 2]
        self.optimal_cut_off = cut_offs[middle_index]
        self.metrics=metrics
        self.cutoffs=cut_offs

        return self

    def predict(self, X):
        grade_predictions = self.model.predict(X)
        class_predictions = (grade_predictions < self.optimal_cut_off).astype(int)
        return grade_predictions, class_predictions

    @property
    def metrics_(self):
        return self.metrics

    @property
    def cutoffs_(self):
        return self.cutoffs

    @property
    def alpha_(self):
        return self.model.alpha_

    @property
    def coef_(self):
        return self.model.coef_

    @property
    def intercept_(self):
        return self.model.intercept_

    @property
    def feature_names_in_(self):
        return getattr(self.model, 'feature_names_in_', None)
